/* SPDX-License-Identifier: GPL-3.0-or-later */
#ifndef FZF_ADDITIONS_H_
#define FZF_ADDITIONS_H_

#include "fzf.h"

/* See fzf-additions.c. */
bool fzf_has_match(const char *text, fzf_pattern_t *pattern);

#endif /* FZF_ADDITIONS_H_ */
